# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import enum
# non-stdlib imports
# local imports
import blobxfer.models.azure
import blobxfer.models.download
import blobxfer.models.options
import blobxfer.models.synccopy
import blobxfer.models.upload
import blobxfer.operations.azure
import blobxfer.operations.crypto
import blobxfer.util


# global defines
_SUPPORTED_YAML_CONFIG_VERSIONS = frozenset((1,))


# enums
class TransferAction(enum.Enum):
    Download = 1,
    Upload = 2,
    Synccopy = 3,


def add_cli_options(cli_options, action):
    # type: (dict, str) -> None
    """Adds CLI options to the configuration object
    :param dict cli_options: CLI options dict
    :param TransferAction action: action
    """
    cli_options['_action'] = action.name.lower()
    # if url is present, convert to constituent options
    if blobxfer.util.is_not_empty(cli_options.get('storage_url')):
        if (blobxfer.util.is_not_empty(cli_options.get('storage_account')) or
                blobxfer.util.is_not_empty(cli_options.get('mode')) or
                blobxfer.util.is_not_empty(cli_options.get('endpoint')) or
                blobxfer.util.is_not_empty(cli_options.get('remote_path'))):
            raise ValueError(
                'Specified both --storage-url and --storage-account, '
                '--mode, --endpoint, or --remote-path')
        cli_options['storage_account'], mode, \
            cli_options['endpoint'], \
            cli_options['remote_path'], \
            sas = blobxfer.util.explode_azure_storage_url(
                cli_options['storage_url'])
        if blobxfer.util.is_not_empty(sas):
            if blobxfer.util.is_not_empty(cli_options['sas']):
                raise ValueError(
                    'Specified both --storage-url with a SAS token and --sas')
            cli_options['sas'] = sas
        if mode != 'blob' and mode != 'file':
            raise ValueError(
                'Invalid derived mode from --storage-url: {}'.format(mode))
        if mode == 'file':
            cli_options['mode'] = mode
        del mode
        del sas
    storage_account = cli_options.get('storage_account')
    azstorage = {
        'endpoint': cli_options.get('endpoint')
    }
    if blobxfer.util.is_not_empty(storage_account):
        azstorage['accounts'] = {
            storage_account: (
                cli_options.get('access_key') or cli_options.get('sas')
            )
        }
    sa_rp = {
        storage_account: cli_options.get('remote_path')
    }
    local_resource = cli_options.get('local_resource')
    # construct "argument" from cli options
    if action == TransferAction.Download:
        arg = {
            'source': [sa_rp] if sa_rp[storage_account] is not None else None,
            'destination': local_resource if local_resource is not None else
            None,
            'include': cli_options.get('include'),
            'exclude': cli_options.get('exclude'),
            'options': {
                'check_file_md5': cli_options.get('file_md5'),
                'chunk_size_bytes': cli_options.get('chunk_size_bytes'),
                'delete_extraneous_destination': cli_options.get('delete'),
                'delete_only': cli_options.get('delete_only'),
                'max_single_object_concurrency': cli_options.get(
                    'max_single_object_concurrency'),
                'mode': cli_options.get('mode'),
                'overwrite': cli_options.get('overwrite'),
                'recursive': cli_options.get('recursive'),
                'rename': cli_options.get('rename'),
                'rsa_private_key': cli_options.get('rsa_private_key'),
                'rsa_private_key_passphrase': cli_options.get(
                    'rsa_private_key_passphrase'),
                'restore_file_properties': {
                    'attributes': cli_options.get('file_attributes'),
                    'lmt': cli_options.get('restore_file_lmt'),
                    'md5': None,
                },
                'strip_components': cli_options.get('strip_components'),
                'skip_on': {
                    'filesize_match': cli_options.get(
                        'skip_on_filesize_match'),
                    'lmt_ge': cli_options.get('skip_on_lmt_ge'),
                    'md5_match': cli_options.get('skip_on_md5_match'),
                },
            },
        }
    elif action == TransferAction.Synccopy:
        # if url is present, convert to constituent options
        if blobxfer.util.is_not_empty(
                cli_options.get('sync_copy_dest_storage_url')):
            if (blobxfer.util.is_not_empty(
                    cli_options.get('sync_copy_dest_storage_account')) or
                    blobxfer.util.is_not_empty(
                        cli_options.get('sync_copy_dest_mode')) or
                    blobxfer.util.is_not_empty(
                        cli_options.get('sync_copy_dest_remote_path'))):
                raise ValueError(
                    'Specified both --sync-copy-dest-storage-url and '
                    '--sync-copy-dest-storage-account, '
                    '--sync-copy-dest-mode, or'
                    '--sync-copy-dest-remote-path')
            cli_options['sync_copy_dest_storage_account'], mode, _, \
                cli_options['sync_copy_dest_remote_path'], sas = \
                blobxfer.util.explode_azure_storage_url(
                    cli_options['sync_copy_dest_storage_url'])
            if blobxfer.util.is_not_empty(sas):
                if blobxfer.util.is_not_empty(
                        cli_options['sync_copy_dest_sas']):
                    raise ValueError(
                        'Specified both --sync-copy-dest-storage-url with '
                        'a SAS token and --sync-copy-dest-sas')
                cli_options['sync_copy_dest_sas'] = sas
            if mode != 'blob' and mode != 'file':
                raise ValueError(
                    'Invalid derived destination mode from '
                    '--sync-copy-dest-storage-url: {}'.format(mode))
            if mode == 'file':
                cli_options['dest_mode'] = mode
            del mode
            del sas
        sync_copy_dest_storage_account = cli_options.get(
            'sync_copy_dest_storage_account')
        sync_copy_dest_remote_path = cli_options.get(
            'sync_copy_dest_remote_path')
        if (sync_copy_dest_storage_account is not None and
                sync_copy_dest_remote_path is not None):
            sync_copy_dest = [
                {
                    sync_copy_dest_storage_account:
                    sync_copy_dest_remote_path
                }
            ]
            if 'accounts' not in azstorage:
                azstorage['accounts'] = {}
            azstorage['accounts'][sync_copy_dest_storage_account] = (
                cli_options.get('sync_copy_dest_access_key') or
                cli_options.get('sync_copy_dest_sas')
            )
        else:
            sync_copy_dest = None
        arg = {
            'destination': sync_copy_dest,
            'include': cli_options.get('include'),
            'exclude': cli_options.get('exclude'),
            'options': {
                'access_tier': cli_options.get('access_tier'),
                'chunk_size_bytes': cli_options.get('chunk_size_bytes'),
                'delete_extraneous_destination': cli_options.get('delete'),
                'delete_only': cli_options.get('delete_only'),
                'dest_mode': cli_options.get('sync_copy_dest_mode'),
                'mode': cli_options.get('mode'),
                'overwrite': cli_options.get('overwrite'),
                'rename': cli_options.get('rename'),
                'server_side_copy': cli_options.get('server_side_copy'),
                'skip_on': {
                    'filesize_match': cli_options.get(
                        'skip_on_filesize_match'),
                    'lmt_ge': cli_options.get('skip_on_lmt_ge'),
                    'md5_match': cli_options.get('skip_on_md5_match'),
                },
                'strip_components': cli_options.get('strip_components'),
            },
        }
        if storage_account is not None:
            arg['source'] = (
                [sa_rp] if sa_rp[storage_account] is not None else None
            )
        else:
            src_url = cli_options.get('remote_path')
            if blobxfer.util.is_not_empty(src_url):
                arg['source'] = [{
                    '*': src_url
                }]
            else:
                arg['source'] = None
    elif action == TransferAction.Upload:
        arg = {
            'source': [local_resource] if local_resource is not None else None,
            'destination': [sa_rp] if sa_rp[storage_account] is not None else
            None,
            'include': cli_options.get('include'),
            'exclude': cli_options.get('exclude'),
            'options': {
                'access_tier': cli_options.get('access_tier'),
                'chunk_size_bytes': cli_options.get('chunk_size_bytes'),
                'delete_extraneous_destination': cli_options.get('delete'),
                'delete_only': cli_options.get('delete_only'),
                'mode': cli_options.get('mode'),
                'one_shot_bytes': cli_options.get('one_shot_bytes'),
                'overwrite': cli_options.get('overwrite'),
                'recursive': cli_options.get('recursive'),
                'rename': cli_options.get('rename'),
                'rsa_private_key': cli_options.get('rsa_private_key'),
                'rsa_private_key_passphrase': cli_options.get(
                    'rsa_private_key_passphrase'),
                'rsa_public_key': cli_options.get('rsa_public_key'),
                'skip_on': {
                    'filesize_match': cli_options.get(
                        'skip_on_filesize_match'),
                    'lmt_ge': cli_options.get('skip_on_lmt_ge'),
                    'md5_match': cli_options.get('skip_on_md5_match'),
                },
                'stdin_as_page_blob_size': cli_options.get(
                    'stdin_as_page_blob_size'),
                'store_file_properties': {
                    'attributes': cli_options.get('file_attributes'),
                    'cache_control': cli_options.get('file_cache_control'),
                    'content_type': cli_options.get('file_content_type'),
                    'lmt': None,
                    'md5': cli_options.get('file_md5'),
                },
                'strip_components': cli_options.get('strip_components'),
                'vectored_io': {
                    'stripe_chunk_size_bytes': cli_options.get(
                        'stripe_chunk_size_bytes'),
                    'distribution_mode': cli_options.get('distribution_mode'),
                },
            },
        }
    count = 0
    if arg['source'] is None:
        arg.pop('source')
        count += 1
    if arg['destination'] is None:
        arg.pop('destination')
        count += 1
    if count == 1:
        if action == TransferAction.Synccopy:
            raise ValueError(
                '--remote-path and --sync-copy-dest-remote-path must be '
                'specified together through the commandline')
        else:
            raise ValueError(
                '--local-path and --remote-path must be specified together '
                'through the commandline')
    if 'accounts' in azstorage:
        cli_options['azure_storage'] = azstorage
    cli_options[action.name.lower()] = arg


def _merge_setting(cli_options, conf, name, name_cli=None, default=None):
    # type: (dict, dict, str, str, Any) -> Any
    """Merge a setting, preferring the CLI option if set
    :param dict cli_options: cli options
    :param dict conf: configuration sub-block
    :param str name: key name
    :param str name_cli: override key name for cli_options
    :param Any default: default value to set if missing
    :rtype: Any
    :return: merged setting value
    """
    val = cli_options.get(name_cli or name)
    if val is None:
        val = conf.get(name, default)
    return val


def merge_global_settings(config, cli_options):
    # type: (dict, dict) -> None
    """Merge "global" CLI options into main config
    :param dict config: config dict
    :param dict cli_options: cli options
    """
    # check for valid version from YAML
    if (not blobxfer.util.is_none_or_empty(config) and
            ('version' not in config or
             config['version'] not in _SUPPORTED_YAML_CONFIG_VERSIONS)):
        raise ValueError('"version" not specified in YAML config or invalid')
    # get action
    action = cli_options['_action']
    if (action != TransferAction.Upload.name.lower() and
            action != TransferAction.Download.name.lower() and
            action != TransferAction.Synccopy.name.lower()):
        raise ValueError('invalid action: {}'.format(action))
    # merge credentials
    if 'azure_storage' in cli_options:
        if 'azure_storage' not in config:
            config['azure_storage'] = {}
        config['azure_storage'] = blobxfer.util.merge_dict(
            config['azure_storage'], cli_options['azure_storage'])
    if ('azure_storage' not in config or
            blobxfer.util.is_none_or_empty(config['azure_storage'])):
        raise ValueError('azure storage settings not specified')
    # create action options
    if action not in config:
        config[action] = []
    # append full specs, if they exist
    if action in cli_options:
        if 'source' in cli_options[action]:
            srcdst = {
                'source': cli_options[action]['source'],
                'destination': cli_options[action]['destination'],
            }
            cli_options[action].pop('source')
            cli_options[action].pop('destination')
            config[action].append(srcdst)
    # merge general and concurrency options
    if 'options' not in config:
        config['options'] = {}
    if 'concurrency' not in config['options']:
        config['options']['concurrency'] = {}
    if 'timeout' not in config['options']:
        config['options']['timeout'] = {}
    if 'proxy' not in config['options']:
        config['options']['proxy'] = {}
    options = {
        'enable_azure_storage_logger': _merge_setting(
            cli_options, config['options'], 'enable_azure_storage_logger'),
        'log_file': _merge_setting(cli_options, config['options'], 'log_file'),
        'progress_bar': _merge_setting(
            cli_options, config['options'], 'progress_bar', default=True),
        'resume_file': _merge_setting(
            cli_options, config['options'], 'resume_file'),
        'timeout': {
            # TODO deprecated timeout setting
            'timeout': _merge_setting(
                cli_options, config['options']['timeout'], 'timeout',
                name_cli='timeout'),
            'connect': _merge_setting(
                cli_options, config['options']['timeout'], 'connect',
                name_cli='connect_timeout'),
            'read': _merge_setting(
                cli_options, config['options']['timeout'], 'read',
                name_cli='read_timeout'),
            'max_retries': _merge_setting(
                cli_options, config['options']['timeout'], 'max_retries',
                default=1000),
        },
        'verbose': _merge_setting(
            cli_options, config['options'], 'verbose', default=False),
        'quiet': _merge_setting(
            cli_options, config['options'], 'quiet', default=False),
        'dry_run': _merge_setting(
            cli_options, config['options'], 'dry_run', default=False),
        'concurrency': {
            'crypto_processes': _merge_setting(
                cli_options, config['options']['concurrency'],
                'crypto_processes', default=0),
            'disk_threads': _merge_setting(
                cli_options, config['options']['concurrency'],
                'disk_threads', default=0),
            'md5_processes': _merge_setting(
                cli_options, config['options']['concurrency'],
                'md5_processes', default=0),
            'transfer_threads': _merge_setting(
                cli_options, config['options']['concurrency'],
                'transfer_threads', default=0),
        },
        'proxy': {
            'host': _merge_setting(
                cli_options, config['options']['proxy'], 'host',
                name_cli='proxy_host'),
            'username': _merge_setting(
                cli_options, config['options']['proxy'], 'username',
                name_cli='proxy_username'),
            'password': _merge_setting(
                cli_options, config['options']['proxy'], 'password',
                name_cli='proxy_password'),
        }
    }
    config['options'] = options


def create_azure_storage_credentials(config, general_options):
    # type: (dict, blobxfer.models.options.General) ->
    #        blobxfer.operations.azure.StorageCredentials
    """Create an Azure StorageCredentials object from configuration
    :param dict config: config dict
    :param blobxfer.models.options.General: general options
    :rtype: blobxfer.operations.azure.StorageCredentials
    :return: credentials object
    """
    creds = blobxfer.operations.azure.StorageCredentials(general_options)
    endpoint = config['azure_storage'].get('endpoint') or 'core.windows.net'
    for name in config['azure_storage']['accounts']:
        key = config['azure_storage']['accounts'][name]
        creds.add_storage_account(name, key, endpoint)
    return creds


def create_general_options(config, action):
    # type: (dict, TransferAction) -> blobxfer.models.options.General
    """Create a General Options object from configuration
    :param dict config: config dict
    :param TransferAction action: transfer action
    :rtype: blobxfer.models.options.General
    :return: general options object
    """
    conc = config['options']['concurrency']
    # split http proxy host into host:port
    proxy = None
    if blobxfer.util.is_not_empty(config['options']['proxy']['host']):
        tmp = config['options']['proxy']['host'].split(':')
        if len(tmp) != 2:
            raise ValueError('Proxy host is malformed: host should be ip:port')
        username = config['options']['proxy']['username']
        if blobxfer.util.is_none_or_empty(username):
            username = None
        password = config['options']['proxy']['password']
        if blobxfer.util.is_none_or_empty(password):
            password = None
        proxy = blobxfer.models.options.HttpProxy(
            host=tmp[0],
            port=int(tmp[1]),
            username=username,
            password=password,
        )
    return blobxfer.models.options.General(
        concurrency=blobxfer.models.options.Concurrency(
            crypto_processes=conc['crypto_processes'],
            disk_threads=conc['disk_threads'],
            md5_processes=conc['md5_processes'],
            transfer_threads=conc['transfer_threads'],
            action=action.value[0],
        ),
        log_file=config['options']['log_file'],
        progress_bar=config['options']['progress_bar'],
        resume_file=config['options']['resume_file'],
        # TODO deprecated timeout setting
        timeout=blobxfer.models.options.Timeout(
            connect=(
                config['options']['timeout']['connect'] or
                config['options']['timeout']['timeout']
            ),
            read=(
                config['options']['timeout']['read'] or
                config['options']['timeout']['timeout']
            ),
            max_retries=config['options']['timeout']['max_retries'],
        ),
        verbose=config['options']['verbose'],
        quiet=config['options']['quiet'],
        dry_run=config['options']['dry_run'],
        proxy=proxy,
    )


def create_download_specifications(ctx_cli_options, config):
    # type: (dict, dict) -> List[blobxfer.models.download.Specification]
    """Create a list of Download Specification objects from configuration
    :param dict ctx_cli_options: cli options
    :param dict config: config dict
    :rtype: list
    :return: list of Download Specification objects
    """
    cli_conf = ctx_cli_options[ctx_cli_options['_action']]
    cli_options = cli_conf['options']
    specs = []
    for conf in config['download']:
        if 'options' in conf:
            conf_options = conf['options']
        else:
            conf_options = {}
        # create download options
        mode = _merge_setting(
            cli_options, conf_options, 'mode', default='auto').lower()
        if mode == 'auto':
            mode = blobxfer.models.azure.StorageModes.Auto
        elif mode == 'append':
            mode = blobxfer.models.azure.StorageModes.Append
        elif mode == 'block':
            mode = blobxfer.models.azure.StorageModes.Block
        elif mode == 'file':
            mode = blobxfer.models.azure.StorageModes.File
        elif mode == 'page':
            mode = blobxfer.models.azure.StorageModes.Page
        else:
            raise ValueError('unknown mode: {}'.format(mode))
        # load RSA private key PEM file if specified
        rpk = _merge_setting(
            cli_options, conf_options, 'rsa_private_key', default=None)
        if blobxfer.util.is_not_empty(rpk):
            rpkp = _merge_setting(
                cli_options, conf_options, 'rsa_private_key_passphrase',
                default=None)
            rpk = blobxfer.operations.crypto.load_rsa_private_key_file(
                rpk, rpkp)
        else:
            rpk = None
        # create specification
        conf_sod = conf_options.get('skip_on', {})
        cli_sod = cli_options['skip_on']
        conf_rfp = conf_options.get('restore_file_properties', {})
        cli_rfp = cli_options['restore_file_properties']
        ds = blobxfer.models.download.Specification(
            download_options=blobxfer.models.options.Download(
                check_file_md5=_merge_setting(
                    cli_options, conf_options, 'check_file_md5',
                    default=False),
                chunk_size_bytes=_merge_setting(
                    cli_options, conf_options, 'chunk_size_bytes',
                    default=0),
                delete_extraneous_destination=_merge_setting(
                    cli_options, conf_options,
                    'delete_extraneous_destination', default=False),
                delete_only=_merge_setting(
                    cli_options, conf_options, 'delete_only', default=False),
                max_single_object_concurrency=_merge_setting(
                    cli_options, conf_options,
                    'max_single_object_concurrency', default=8),
                mode=mode,
                overwrite=_merge_setting(
                    cli_options, conf_options, 'overwrite', default=True),
                recursive=_merge_setting(
                    cli_options, conf_options, 'recursive', default=True),
                rename=_merge_setting(
                    cli_options, conf_options, 'rename', default=False),
                restore_file_properties=blobxfer.models.options.FileProperties(
                    attributes=_merge_setting(
                        cli_rfp, conf_rfp, 'attributes',
                        default=False),
                    cache_control=None,
                    content_type=None,
                    lmt=_merge_setting(
                        cli_rfp, conf_rfp, 'lmt', default=False),
                    md5=None,
                ),
                rsa_private_key=rpk,
                strip_components=_merge_setting(
                    cli_options, conf_options, 'strip_components',
                    default=0),
            ),
            skip_on_options=blobxfer.models.options.SkipOn(
                filesize_match=_merge_setting(
                    cli_sod, conf_sod, 'filesize_match', default=False),
                lmt_ge=_merge_setting(
                    cli_sod, conf_sod, 'lmt_ge', default=False),
                md5_match=_merge_setting(
                    cli_sod, conf_sod, 'md5_match', default=False),
            ),
            local_destination_path=blobxfer.models.download.
            LocalDestinationPath(
                conf['destination']
            )
        )
        if (ds.options.delete_only and
                not ds.options.delete_extraneous_destination):
            raise ValueError('delete only specified without delete')
        # create remote source paths
        for src in conf['source']:
            if len(src) != 1:
                raise RuntimeError(
                    'invalid number of source pairs specified per entry')
            sa = next(iter(src))
            asp = blobxfer.operations.azure.SourcePath()
            asp.add_path_with_storage_account(src[sa], sa)
            incl = _merge_setting(cli_conf, conf, 'include', default=None)
            if blobxfer.util.is_not_empty(incl):
                asp.add_includes(incl)
            excl = _merge_setting(cli_conf, conf, 'exclude', default=None)
            if blobxfer.util.is_not_empty(excl):
                asp.add_excludes(excl)
            ds.add_azure_source_path(asp)
        # append spec to list
        specs.append(ds)
    return specs


def create_synccopy_specifications(ctx_cli_options, config):
    # type: (dict, dict) -> List[blobxfer.models.synccopy.Specification]
    """Create a list of SyncCopy Specification objects from configuration
    :param dict ctx_cli_options: cli options
    :param dict config: config dict
    :rtype: list
    :return: list of SyncCopy Specification objects
    """
    cli_conf = ctx_cli_options[ctx_cli_options['_action']]
    cli_options = cli_conf['options']
    specs = []
    for conf in config['synccopy']:
        if 'options' in conf:
            conf_options = conf['options']
        else:
            conf_options = {}
        # get source mode
        mode = _merge_setting(
            cli_options, conf_options, 'mode', default='auto').lower()
        if mode == 'auto':
            mode = blobxfer.models.azure.StorageModes.Auto
        elif mode == 'append':
            mode = blobxfer.models.azure.StorageModes.Append
        elif mode == 'block':
            mode = blobxfer.models.azure.StorageModes.Block
        elif mode == 'file':
            mode = blobxfer.models.azure.StorageModes.File
        elif mode == 'page':
            mode = blobxfer.models.azure.StorageModes.Page
        else:
            raise ValueError('unknown source mode: {}'.format(mode))
        # get destination mode
        destmode = _merge_setting(
            cli_options, conf_options, 'dest_mode', name_cli='dest_mode')
        if blobxfer.util.is_none_or_empty(destmode):
            destmode = mode
        else:
            destmode = destmode.lower()
            if destmode == 'auto':
                destmode = blobxfer.models.azure.StorageModes.Auto
            elif destmode == 'append':
                destmode = blobxfer.models.azure.StorageModes.Append
            elif destmode == 'block':
                destmode = blobxfer.models.azure.StorageModes.Block
            elif destmode == 'file':
                destmode = blobxfer.models.azure.StorageModes.File
            elif destmode == 'page':
                destmode = blobxfer.models.azure.StorageModes.Page
            else:
                raise ValueError('unknown dest mode: {}'.format(destmode))
        # create specification
        conf_sod = conf_options.get('skip_on', {})
        cli_sod = cli_options['skip_on']
        scs = blobxfer.models.synccopy.Specification(
            synccopy_options=blobxfer.models.options.SyncCopy(
                access_tier=_merge_setting(
                    cli_options, conf_options, 'access_tier', default=None),
                delete_extraneous_destination=_merge_setting(
                    cli_options, conf_options,
                    'delete_extraneous_destination', default=False),
                delete_only=_merge_setting(
                    cli_options, conf_options, 'delete_only', default=False),
                dest_mode=destmode,
                mode=mode,
                overwrite=_merge_setting(
                    cli_options, conf_options, 'overwrite', default=True),
                recursive=_merge_setting(
                    cli_options, conf_options, 'recursive', default=True),
                rename=_merge_setting(
                    cli_options, conf_options, 'rename', default=False),
                server_side_copy=_merge_setting(
                    cli_options, conf_options, 'server_side_copy',
                    default=True),
                strip_components=_merge_setting(
                    cli_options, conf_options, 'strip_components',
                    default=0),
            ),
            skip_on_options=blobxfer.models.options.SkipOn(
                filesize_match=_merge_setting(
                    cli_sod, conf_sod, 'filesize_match', default=False),
                lmt_ge=_merge_setting(
                    cli_sod, conf_sod, 'lmt_ge', default=False),
                md5_match=_merge_setting(
                    cli_sod, conf_sod, 'md5_match', default=False),
            ),
        )
        if (scs.options.delete_only and
                not scs.options.delete_extraneous_destination):
            raise ValueError('delete only specified without delete')
        # create remote source paths
        for src in conf['source']:
            sa = next(iter(src))
            asp = blobxfer.operations.azure.SourcePath()
            if sa != '*':
                asp.add_path_with_storage_account(src[sa], sa)
                incl = _merge_setting(cli_conf, conf, 'include', default=None)
                if blobxfer.util.is_not_empty(incl):
                    asp.add_includes(incl)
                excl = _merge_setting(cli_conf, conf, 'exclude', default=None)
                if blobxfer.util.is_not_empty(excl):
                    asp.add_excludes(excl)
            else:
                if not scs.options.server_side_copy:
                    raise ValueError(
                        'Server side copy must be enabled for abitrary '
                        'source remote paths')
                asp.add_arbitrary_remote_url(src[sa])
            scs.add_azure_source_path(asp)
        # create remote destination paths
        for dst in conf['destination']:
            if len(dst) != 1:
                raise RuntimeError(
                    'invalid number of destination pairs specified per entry')
            sa = next(iter(dst))
            adp = blobxfer.operations.azure.DestinationPath()
            adp.add_path_with_storage_account(dst[sa], sa)
            scs.add_azure_destination_path(adp)
        # append spec to list
        specs.append(scs)
    return specs


def create_upload_specifications(ctx_cli_options, config):
    # type: (dict, dict) -> List[blobxfer.models.upload.Specification]
    """Create a list of Upload Specification objects from configuration
    :param dict ctx_cli_options: cli options
    :param dict config: config dict
    :rtype: list
    :return: list of Upload Specification objects
    """
    cli_conf = ctx_cli_options[ctx_cli_options['_action']]
    cli_options = cli_conf['options']
    specs = []
    for conf in config['upload']:
        if 'options' in conf:
            conf_options = conf['options']
        else:
            conf_options = {}
        # create upload options
        mode = _merge_setting(
            cli_options, conf_options, 'mode', default='auto').lower()
        if mode == 'auto':
            mode = blobxfer.models.azure.StorageModes.Auto
        elif mode == 'append':
            mode = blobxfer.models.azure.StorageModes.Append
        elif mode == 'block':
            mode = blobxfer.models.azure.StorageModes.Block
        elif mode == 'file':
            mode = blobxfer.models.azure.StorageModes.File
        elif mode == 'page':
            mode = blobxfer.models.azure.StorageModes.Page
        else:
            raise ValueError('unknown mode: {}'.format(mode))
        # load RSA public key PEM if specified
        rpk = _merge_setting(cli_options, conf_options, 'rsa_public_key')
        if blobxfer.util.is_not_empty(rpk):
            rpk = blobxfer.operations.crypto.load_rsa_public_key_file(rpk)
        if rpk is None:
            # load RSA private key PEM file if specified
            rpk = _merge_setting(
                cli_options, conf_options, 'rsa_private_key')
            if blobxfer.util.is_not_empty(rpk):
                rpkp = _merge_setting(
                    cli_options, conf_options, 'rsa_private_key_passphrase')
                rpk = blobxfer.operations.crypto.load_rsa_private_key_file(
                    rpk, rpkp)
                rpk = rpk.public_key()
            else:
                rpk = None
        # create local source paths
        lsp = blobxfer.models.upload.LocalSourcePath()
        lsp.add_paths(conf['source'])
        incl = _merge_setting(cli_conf, conf, 'include', default=None)
        if blobxfer.util.is_not_empty(incl):
            lsp.add_includes(incl)
        excl = _merge_setting(cli_conf, conf, 'exclude', default=None)
        if blobxfer.util.is_not_empty(excl):
            lsp.add_excludes(excl)
        # create specification
        conf_sfp = conf_options.get('store_file_properties', {})
        cli_sfp = cli_options['store_file_properties']
        conf_vio = conf_options.get('vectored_io', {})
        cli_vio = cli_options['vectored_io']
        conf_sod = conf_options.get('skip_on', {})
        cli_sod = cli_options['skip_on']
        us = blobxfer.models.upload.Specification(
            upload_options=blobxfer.models.options.Upload(
                access_tier=_merge_setting(
                    cli_options, conf_options, 'access_tier', default=None),
                chunk_size_bytes=_merge_setting(
                    cli_options, conf_options, 'chunk_size_bytes',
                    default=0),
                delete_extraneous_destination=_merge_setting(
                    cli_options, conf_options,
                    'delete_extraneous_destination', default=False),
                delete_only=_merge_setting(
                    cli_options, conf_options, 'delete_only', default=False),
                mode=mode,
                one_shot_bytes=_merge_setting(
                    cli_options, conf_options, 'one_shot_bytes', default=0),
                overwrite=_merge_setting(
                    cli_options, conf_options, 'overwrite', default=True),
                recursive=_merge_setting(
                    cli_options, conf_options, 'recursive', default=True),
                rename=_merge_setting(
                    cli_options, conf_options, 'rename', default=False),
                rsa_public_key=rpk,
                store_file_properties=blobxfer.models.options.FileProperties(
                    attributes=_merge_setting(
                        cli_sfp, conf_sfp, 'attributes', default=False),
                    cache_control=_merge_setting(
                        cli_sfp, conf_sfp, 'cache_control', default=None),
                    content_type=_merge_setting(
                        cli_sfp, conf_sfp, 'content_type', default=None),
                    lmt=None,
                    md5=_merge_setting(
                        cli_sfp, conf_sfp, 'md5', default=False),
                ),
                stdin_as_page_blob_size=_merge_setting(
                    cli_options, conf_options, 'stdin_as_page_blob_size',
                    default=0),
                strip_components=_merge_setting(
                    cli_options, conf_options, 'strip_components',
                    default=0),
                vectored_io=blobxfer.models.options.VectoredIo(
                    stripe_chunk_size_bytes=_merge_setting(
                        cli_vio, conf_vio, 'stripe_chunk_size_bytes',
                        default=1073741824),
                    distribution_mode=blobxfer.
                    models.upload.VectoredIoDistributionMode(
                        _merge_setting(
                            cli_vio, conf_vio, 'distribution_mode',
                            default='disabled').lower()),
                ),
            ),
            skip_on_options=blobxfer.models.options.SkipOn(
                filesize_match=_merge_setting(
                    cli_sod, conf_sod, 'filesize_match', default=False),
                lmt_ge=_merge_setting(
                    cli_sod, conf_sod, 'lmt_ge', default=False),
                md5_match=_merge_setting(
                    cli_sod, conf_sod, 'md5_match', default=False),
            ),
            local_source_path=lsp,
        )
        if (us.options.delete_only and
                not us.options.delete_extraneous_destination):
            raise ValueError('delete only specified without delete')
        # create remote destination paths
        for dst in conf['destination']:
            if len(dst) != 1:
                raise RuntimeError(
                    'invalid number of destination pairs specified per entry')
            sa = next(iter(dst))
            adp = blobxfer.operations.azure.DestinationPath()
            adp.add_path_with_storage_account(dst[sa], sa)
            us.add_azure_destination_path(adp)
        # append spec to list
        specs.append(us)
    return specs



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import absolute_import, division, print_function
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip
)
# stdlib imports
import base64
import copy
import datetime
import hashlib
import logging
import logging.handlers
import mimetypes
try:
    from os import scandir as scandir
except ImportError:  # noqa
    from scandir import scandir as scandir
import platform
import re
import sys
# non-stdlib imports
import dateutil.parser
import dateutil.tz
import future.utils
# local imports

# global defines
MEGABYTE = 1048576
_DEFAULT_MIME_TYPE = 'application/octet-stream'
_ON_LINUX = platform.system() == 'Linux'
_ON_WINDOWS = platform.system() == 'Windows'
_REGISTERED_LOGGER_HANDLERS = []
_PAGEBLOB_BOUNDARY = 512


def on_python2():
    # type: (None) -> bool
    """Execution on python2
    :rtype: bool
    :return: if on Python2
    """
    return future.utils.PY2


def on_linux():  # noqa
    # type: (None) -> bool
    """Execution on Linux
    :rtype: bool
    :return: if on Linux
    """
    return _ON_LINUX


def on_windows():  # noqa
    # type: (None) -> bool
    """Execution on Windows
    :rtype: bool
    :return: if on Windows
    """
    return _ON_WINDOWS


def py_ver():  # noqa
    # type: (None) -> Tuple[int, int]
    """Python version
    :rtype tuple
    :return: version info tuple
    """
    return sys.version_info


def setup_logger(logger, logfile):  # noqa
    # type: (logger, str) -> None
    """Set up logger"""
    global _REGISTERED_LOGGER_HANDLERS
    logger.setLevel(logging.DEBUG)
    if is_none_or_empty(logfile):
        handler = logging.StreamHandler()
    else:
        handler = logging.FileHandler(logfile, encoding='utf-8')
    logging.getLogger().addHandler(handler)
    formatter = logging.Formatter('%(asctime)s %(levelname)s - %(message)s')
    formatter.default_msec_format = '%s.%03d'
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False
    _REGISTERED_LOGGER_HANDLERS.append(handler)


def set_verbose_logger_handlers():  # noqa
    # type: (None) -> None
    """Set logger handler formatters to more detail"""
    global _REGISTERED_LOGGER_HANDLERS
    formatter = logging.Formatter(
        '%(asctime)s %(levelname)s %(name)s:%(funcName)s:%(lineno)d '
        '%(message)s')
    formatter.default_msec_format = '%s.%03d'
    for handler in _REGISTERED_LOGGER_HANDLERS:
        handler.setFormatter(formatter)


def is_none_or_empty(obj):
    # type: (any) -> bool
    """Determine if object is None or empty
    :type any obj: object
    :rtype: bool
    :return: if object is None or empty
    """
    return obj is None or len(obj) == 0


def is_not_empty(obj):
    # type: (any) -> bool
    """Determine if object is not None and is length is > 0
    :type any obj: object
    :rtype: bool
    :return: if object is not None and length is > 0
    """
    return obj is not None and len(obj) > 0


def join_thread(thr):
    # type: (threading.Thread) -> None
    """Join a thread
    :type threading.Thread thr: thread to join
    """
    if on_python2():
        while True:
            thr.join(timeout=1)
            if not thr.isAlive():
                break
    else:
        thr.join()


def merge_dict(dict1, dict2):
    # type: (dict, dict) -> dict
    """Recursively merge dictionaries: dict2 on to dict1. This differs
    from dict.update() in that values that are dicts are recursively merged.
    Note that only dict value types are merged, not lists, etc.

    :param dict dict1: dictionary to merge to
    :param dict dict2: dictionary to merge with
    :rtype: dict
    :return: merged dictionary
    """
    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        raise ValueError('dict1 or dict2 is not a dictionary')
    result = copy.deepcopy(dict1)
    for k, v in dict2.items():
        if k in result and isinstance(result[k], dict):
            result[k] = merge_dict(result[k], v)
        else:
            result[k] = copy.deepcopy(v)
    return result


def datetime_now():
    # type: (None) -> datetime.datetime
    """Return a timezone-aware datetime instance with local offset
    :rtype: datetime.datetime
    :return: datetime now with local tz
    """
    return datetime.datetime.now(tz=dateutil.tz.tzlocal())


def datetime_from_timestamp(ts, tz=None, as_utc=False):
    # type: (float, dateutil.tz, bool) -> datetime.datetime
    """Convert a timestamp into datetime with offset
    :param float ts: timestamp
    :param dateutil.tz tz: time zone or local tz if not specified
    :param bool as_utc: convert datetime to UTC
    :rtype: datetime.datetime
    :return: converted timestamp to datetime
    """
    if tz is None:
        tz = dateutil.tz.tzlocal()
    dt = datetime.datetime.fromtimestamp(ts, tz=tz)
    if as_utc:
        return dt.astimezone(tz=dateutil.tz.tzutc())
    else:
        return dt


def scantree(path):
    # type: (str) -> os.DirEntry
    """Recursively scan a directory tree
    :param str path: path to scan
    :rtype: DirEntry
    :return: DirEntry via generator
    """
    for entry in scandir(path):
        if entry.is_dir(follow_symlinks=True):
            # due to python2 compat, cannot use yield from here
            for t in scantree(entry.path):
                yield t
        else:
            yield entry


def get_mime_type(filename):
    # type: (str) -> str
    """Guess the type of a file based on its filename
    :param str filename: filename to guess the content-type
    :rtype: str
    :rturn: string of form 'class/type' for MIME content-type header
    """
    # TODO extract encoding for ContentEncoding property
    return mimetypes.guess_type(filename)[0] or _DEFAULT_MIME_TYPE


def base64_encode_as_string(obj):  # noqa
    # type: (any) -> str
    """Encode object to base64
    :param any obj: object to encode
    :rtype: str
    :return: base64 encoded string
    """
    if on_python2():
        return base64.b64encode(obj)
    else:
        return str(base64.b64encode(obj), 'ascii')


def base64_decode_string(string):
    # type: (str) -> str
    """Base64 decode a string
    :param str string: string to decode
    :rtype: str
    :return: decoded string
    """
    return base64.b64decode(string)


def new_md5_hasher():
    # type: (None) -> md5.MD5
    """Create a new MD5 hasher
    :rtype: md5.MD5
    :return: new MD5 hasher
    """
    return hashlib.md5()


def page_align_content_length(length):
    # type: (int) -> int
    """Compute page boundary alignment
    :param int length: content length
    :rtype: int
    :return: aligned byte boundary
    """
    mod = length % _PAGEBLOB_BOUNDARY
    if mod != 0:
        return length + (_PAGEBLOB_BOUNDARY - mod)
    return length


def normalize_azure_path(path):
    # type: (str) -> str
    """Normalize remote path (strip slashes and use forward slashes)
    :param str path: path to normalize
    :rtype: str
    :return: normalized path
    """
    if is_none_or_empty(path):
        raise ValueError('provided path is invalid')
    _path = path.strip('/').strip('\\')
    return '/'.join(re.split('/|\\\\', _path))


def explode_azure_path(path):
    # type: (str) -> Tuple[str, str]
    """Explodes an azure path into a container or fileshare and the
    remaining virtual path
    :param str path: path to explode
    :rtype: tuple
    :return: container, vpath
    """
    rpath = normalize_azure_path(path).split('/')
    container = str(rpath[0])
    if len(rpath) > 1:
        rpath = '/'.join(rpath[1:])
    else:
        rpath = ''
    return container, rpath


def blob_is_snapshot(url):
    # type: (str) -> bool
    """Checks if the blob is a snapshot blob
    :param url str: blob url
    :rtype: bool
    :return: if blob is a snapshot blob
    """
    if '?snapshot=' in url:
        try:
            dateutil.parser.parse(url.split('?snapshot=')[-1])
            return True
        except (ValueError, OverflowError):
            pass
    return False


def parse_blob_snapshot_parameter(url):
    # type: (str) -> str
    """Retrieves the blob snapshot parameter from a url
    :param url str: blob url
    :rtype: str
    :return: snapshot parameter
    """
    if blob_is_snapshot(url):
        tmp = url.split('?snapshot=')
        if len(tmp) == 2:
            return tmp[0], tmp[1]
    return None


def parse_fileshare_or_file_snapshot_parameter(url):
    # type: (str) -> Tuple[str, str]
    """Checks if the fileshare or file is a snapshot
    :param url str: file url
    :rtype: tuple
    :return: (url, snapshot)
    """
    if is_not_empty(url):
        if '?sharesnapshot=' in url:
            try:
                tmp = url.split('?sharesnapshot=')
                if len(tmp) == 2:
                    dateutil.parser.parse(tmp[1])
                    return tmp[0], tmp[1]
            except (ValueError, OverflowError):
                pass
        elif '?snapshot=' in url:
            try:
                tmp = url.split('?snapshot=')
                if len(tmp) == 2:
                    dateutil.parser.parse(tmp[1])
                    return tmp[0], tmp[1]
            except (ValueError, OverflowError):
                pass
    return url, None


def explode_azure_storage_url(url):
    # type: (str) -> Tuple[str, str, str, str, str]
    """Explode Azure Storage URL into parts
    :param url str: storage url
    :rtype: tuple
    :return: (sa, mode, ep, rpath, sas)
    """
    tmp = url.split('/')
    host = tmp[2].split('.')
    sa = host[0]
    mode = host[1].lower()
    ep = '.'.join(host[2:])
    tmp = '/'.join(tmp[3:]).split('?')
    rpath = tmp[0]
    if len(tmp) > 1:
        sas = tmp[1]
    else:
        sas = None
    return sa, mode, ep, rpath, sas



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import logging
import re
# non-stdlib imports
import requests
# local imports
import blobxfer.models
import blobxfer.models.crypto
import blobxfer.models.metadata
import blobxfer.models.options
import blobxfer.operations.azure.blob.append
import blobxfer.operations.azure.blob.block
import blobxfer.operations.azure.blob.page
import blobxfer.operations.azure.file
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)
# global defines
_METADATA_VIRTUAL_DIRECTORY = 'hdi_isfolder'


class StorageCredentials(object):
    """Azure Storage Credentials"""
    def __init__(self, general_options):
        # type: (StorageCredentials, blobxfer.models.options.General) -> None
        """Ctor for StorageCredentials
        :param StorageCredentials self: this
        :param blobxfer.models.options.General: general options
        """
        self._storage_accounts = {}
        self._general_options = general_options

    def add_storage_account(self, name, key, endpoint):
        # type: (StorageCredentials, str, str, str) -> None
        """Add a storage account
        :param StorageCredentials self: this
        :param str name: name of storage account to store
        :param str key: storage key or sas
        :param str endpoint: endpoint
        """
        if name in self._storage_accounts:
            raise ValueError(
                '{} already exists in storage accounts'.format(name))
        self._storage_accounts[name] = StorageAccount(
            name, key, endpoint,
            self._general_options.concurrency.transfer_threads,
            self._general_options.timeout,
            self._general_options.proxy,
        )

    def get_storage_account(self, name):
        # type: (StorageCredentials, str) -> StorageAccount
        """Get storage account details
        :param StorageCredentials self: this
        :param str name: name of storage account to retrieve
        :rtype: StorageAccount
        :return: storage account details
        """
        return self._storage_accounts[name]


class StorageAccount(object):
    """Azure Storage Account"""
    _VALID_BASE64_RE = re.compile(
        '^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|'
        '[A-Za-z0-9+/]{2}==)$')

    def __init__(self, name, key, endpoint, transfer_threads, timeout, proxy):
        # type: (StorageAccount, str, str, str, int,
        #        blobxfer.models.options.Timeout,
        #        blobxfer.models.options.HttpProxy) -> None
        """Ctor for StorageAccount
        :param str name: name of storage account
        :param str key: storage key or sas
        :param str endpoint: endpoint
        :param int transfer_threads: number of transfer threads
        :param blobxfer.models.options.Timeout timeout: timeout
        :param blobxfer.models.options.HttpProxy proxy: proxy
        """
        if blobxfer.util.is_none_or_empty(key):
            raise ValueError(
                ('no authentication credential given for storage '
                 'account: {}').format(name))
        self._endpoint = None
        self._append_blob_client = None
        self._block_blob_client = None
        self._file_client = None
        self._page_blob_client = None
        self.name = name
        self.key = key
        self.endpoint = endpoint
        self.is_sas = StorageAccount._key_is_sas(self.key)
        self.can_create_containers = self._container_manipulation_allowed()
        self.can_list_container_objects = (
            self._credential_allows_container_list()
        )
        self.can_read_object = self._credential_allows_object_read()
        self.can_write_object = self._credential_allows_object_write()
        if self.is_sas:
            # ensure sas can work on objects
            if not self._ensure_object_manipulation_allowed():
                raise ValueError(
                    'the provided SAS does not allow object-level access for '
                    'account: {}'.format(self.name))
            # normalize sas keys
            if self.key.startswith('?'):
                self.key = self.key[1:]
        else:
            # check if sa shared key is base64
            if StorageAccount._VALID_BASE64_RE.match(self.key) is None:
                raise ValueError(
                    ('specified storage account key is invalid for storage '
                     'account: {}').format(self.name))
        logger.debug(
            'credential: account={} endpoint={} is_sas={} '
            'can_create_containers={} can_list_container_objects={} '
            'can_read_object={} can_write_object={}'.format(
                self.name, self.endpoint, self.is_sas,
                self.can_create_containers, self.can_list_container_objects,
                self.can_read_object, self.can_write_object))
        # create requests session for connection pooling
        self.session = requests.Session()
        self.session.mount(
            'https://',
            requests.adapters.HTTPAdapter(
                pool_connections=transfer_threads,
                pool_maxsize=transfer_threads << 2,
                pool_block=False,
            )
        )
        self._create_clients(timeout, proxy)

    @property
    def endpoint(self):
        # type: (StorageAccount) -> str
        """Get endpoint
        :param StorageAccount self: this
        :rtype: str
        :return: endpoint
        """
        return self._endpoint

    @endpoint.setter
    def endpoint(self, value):
        # type: (StorageAccount, str) -> None
        """Set endpoint
        :param StorageAccount self: this
        :param str value: endpoint
        """
        tmp = value.split('.')
        if (len(tmp) <= 1 or not tmp[0].isalnum()):
            raise ValueError('endpoint is invalid: {}'.format(value))
        self._endpoint = value

    @staticmethod
    def _key_is_sas(key):
        # type: (str) -> bool
        """Determine if key is a sas
        :param str key: key to parse
        :rtype: bool
        :return: if key is a sas
        """
        # keys starting with ? are sas keys as ? is not in the base-64
        # character range
        if key.startswith('?'):
            return True
        else:
            # & is not in the base-64 character range, so technically
            # the presence of this character means the key is a sas. however,
            # perform a stronger check for the sig= parameter.
            tmp = key.split('&')
            if len(tmp) == 1:
                return False
            elif any(x.startswith('sig=') for x in tmp):
                return True
        return False

    def _container_manipulation_allowed(self):
        # type: (StorageAccount) -> bool
        """Check if container manipulation is allowed
        :param StorageAccount self: this
        :rtype: bool
        :return: if container manipulation is allowed
        """
        if self.is_sas:
            # search for account sas "c" resource
            sasparts = self.key.split('&')
            for part in sasparts:
                tmp = part.split('=')
                if tmp[0] == 'srt':
                    return 'c' in tmp[1]
            # this is a sas without the srt parameter, so this must be
            # a service-level sas which doesn't allow container manipulation
            return False
        else:
            # storage account key always allows container manipulation
            return True

    def _ensure_object_manipulation_allowed(self):
        # type: (StorageAccount) -> bool
        """Check if object manipulation is allowed
        :param StorageAccount self: this
        :rtype: bool
        :return: if object manipulation is allowed
        """
        if self.is_sas:
            # search for account sas "o" resource
            sasparts = self.key.split('&')
            for part in sasparts:
                tmp = part.split('=')
                if tmp[0] == 'srt':
                    return 'o' in tmp[1]
            # this is a sas without the srt parameter, so this must be
            # a service-level sas which always allows object manipulation
            return True
        else:
            # storage account key always allows object manipulation
            return True

    def _credential_allows_container_list(self):
        # type: (StorageAccount) -> bool
        """Check if container list is allowed
        :param StorageAccount self: this
        :rtype: bool
        :return: if container list is allowed
        """
        if self.is_sas:
            sasparts = self.key.split('&')
            caccess = self.can_create_containers
            # search for container signed resource for service level sas
            if not caccess:
                for part in sasparts:
                    tmp = part.split('=')
                    if tmp[0] == 'sr':
                        caccess = 'c' in tmp[1] or 's' in tmp[1]
                        break
                    elif tmp[0] == 'si':
                        # assume sas policies allow container list
                        return True
            # search for list permission
            if caccess:
                for part in sasparts:
                    tmp = part.split('=')
                    if tmp[0] == 'sp':
                        return 'l' in tmp[1]
            # sas doesn't allow container level list
            return False
        else:
            # storage account key always allows container list
            return True

    def _credential_allows_object_read(self):
        # type: (StorageAccount) -> bool
        """Check if object read is allowed
        :param StorageAccount self: this
        :rtype: bool
        :return: if object read is allowed
        """
        if self.is_sas:
            # search for read permission
            sasparts = self.key.split('&')
            for part in sasparts:
                tmp = part.split('=')
                if tmp[0] == 'sp':
                    return 'r' in tmp[1]
                elif tmp[0] == 'si':
                    # assume sas policies allow object read
                    return True
            return False
        else:
            # storage account key always allows object read
            return True

    def _credential_allows_object_write(self):
        # type: (StorageAccount) -> bool
        """Check if object write is allowed
        :param StorageAccount self: this
        :rtype: bool
        :return: if object write is allowed
        """
        if self.is_sas:
            # search for write permission
            sasparts = self.key.split('&')
            for part in sasparts:
                tmp = part.split('=')
                if tmp[0] == 'sp':
                    return 'w' in tmp[1]
                elif tmp[0] == 'si':
                    # assume sas policies allow object write
                    return True
            return False
        else:
            # storage account key always allows object write
            return True

    def _create_clients(self, timeout, proxy):
        # type: (StorageAccount, blobxfer.models.options.Timeout,
        #        blobxfer.models.options.HttpProxy) -> None
        """Create Azure Storage clients
        :param StorageAccount self: this
        :param blobxfer.models.options.Timeout timeout: timeout
        :param blobxfer.models.options.HttpProxy proxy: proxy
        """
        self._append_blob_client = \
            blobxfer.operations.azure.blob.append.create_client(
                self, timeout, proxy)
        self._block_blob_client = \
            blobxfer.operations.azure.blob.block.create_client(
                self, timeout, proxy)
        self._file_client = blobxfer.operations.azure.file.create_client(
            self, timeout, proxy)
        self._page_blob_client = \
            blobxfer.operations.azure.blob.page.create_client(
                self, timeout, proxy)

    @property
    def append_blob_client(self):
        # type: (StorageAccount) -> azure.storage.blob.AppendBlobService
        """Get append blob client
        :param StorageAccount self: this
        :rtype: azure.storage.blob.AppendBlobService
        :return: append blob client
        """
        return self._append_blob_client

    @property
    def block_blob_client(self):
        # type: (StorageAccount) -> azure.storage.blob.BlockBlobService
        """Get block blob client
        :param StorageAccount self: this
        :rtype: azure.storage.blob.BlockBlobService
        :return: block blob client
        """
        return self._block_blob_client

    @property
    def file_client(self):
        # type: (StorageAccount) -> azure.storage.file.FileService
        """Get file client
        :param StorageAccount self: this
        :rtype: azure.storage.file.FileService
        :return: file client
        """
        return self._file_client

    @property
    def page_blob_client(self):
        # type: (StorageAccount) -> azure.storage.blob.PageBlobService
        """Get page blob client
        :param StorageAccount self: this
        :rtype: azure.storage.blob.PageBlobService
        :return: page blob client
        """
        return self._page_blob_client


class SourcePath(blobxfer.models._BaseSourcePaths):
    """Azure Source Path"""
    def __init__(self):
        # type: (SourcePath) -> None
        """Ctor for SourcePath
        :param SourcePath self: this
        """
        super().__init__()
        self._path_map = {}

    def add_arbitrary_remote_url(self, remote_path):
        # type: (SourcePath, str) -> None
        """Add an arbitrary remote URL
        :param SourcePath self: this
        :param str remote_path: remote path
        """
        self._paths.append(remote_path)

    def add_path_with_storage_account(self, remote_path, storage_account):
        # type: (SourcePath, str, str) -> None
        """Add a path with an associated storage account
        :param SourcePath self: this
        :param str remote_path: remote path
        :param str storage_account: storage account to associate with path
        """
        if len(self._path_map) >= 1:
            raise RuntimeError(
                'cannot add multiple remote paths to SourcePath objects')
        rpath = blobxfer.util.normalize_azure_path(remote_path)
        self.add_path(rpath)
        self._path_map[rpath] = storage_account

    def lookup_storage_account(self, remote_path):
        # type: (SourcePath, str) -> str
        """Lookup the storage account associated with the remote path
        :param SourcePath self: this
        :param str remote_path: remote path
        :rtype: str
        :return: storage account associated with path
        """
        return self._path_map[blobxfer.util.normalize_azure_path(remote_path)]

    def files(self, creds, options, dry_run):
        # type: (SourcePath, StorageCredentials,
        #        blobxfer.models.options.Download, bool) -> StorageEntity
        """Generator of Azure remote files or blobs
        :param SourcePath self: this
        :param StorageCredentials creds: storage creds
        :param blobxfer.models.options.Download options: download options
        :param bool dry_run: dry run
        :rtype: StorageEntity
        :return: Azure storage entity object
        """
        if options.mode == blobxfer.models.azure.StorageModes.File:
            for file in self._populate_from_list_files(
                    creds, options, dry_run):
                yield file
        else:
            for blob in self._populate_from_list_blobs(
                    creds, options, dry_run):
                yield blob

    def _convert_to_storage_entity_with_encryption_metadata(
            self, options, store_raw_metadata, sa, entity, vio, is_file,
            container, dir, file_snapshot):
        # type: (SourcePath, StorageCredentials, any, bool, StorageAccount,
        #        any, blobxfer.models.metadata.VectoredStripe, bool, str,
        #        str) -> StorageEntity
        """Convert entity into StorageEntity with encryption metadata if avail
        :param SourcePath self: this
        :param StorageCredentials creds: storage creds
        :param object options: download or synccopy options
        :param bool store_raw_metadata: store raw metadata
        :param StorageAccount sa: storage account
        :param object entity: Storage File or Blob object
        :param blobxfer.models.metadata.VectoredStripe vio: Vectored stripe
        :param bool is_file: is a file object
        :param str container: container
        :param str dir: Azure File directory structure
        :rtype: StorageEntity
        :return: Azure storage entity object
        """
        if (not store_raw_metadata and
                blobxfer.models.crypto.EncryptionMetadata.
                encryption_metadata_exists(entity.metadata)):
            ed = blobxfer.models.crypto.EncryptionMetadata()
            ed.convert_from_json(
                entity.metadata, entity.name, options.rsa_private_key)
        else:
            ed = None
        ase = blobxfer.models.azure.StorageEntity(container, ed)
        if is_file:
            ase.populate_from_file(
                sa, entity, dir, vio=vio,
                store_raw_metadata=store_raw_metadata, snapshot=file_snapshot)
        else:
            ase.populate_from_blob(
                sa, entity, vio=vio, store_raw_metadata=store_raw_metadata)
        return ase

    def _handle_vectored_io_stripe(
            self, creds, options, store_raw_metadata, sa, entity, is_file,
            container, dir=None, file_snapshot=None):
        # type: (SourcePath, StorageCredentials, any, bool, StorageAccount,
        #        any, bool, str, str) -> StorageEntity
        """Handle Vectored IO stripe entries
        :param SourcePath self: this
        :param StorageCredentials creds: storage creds
        :param object options: download or synccopy options
        :param bool store_raw_metadata: store raw metadata
        :param StorageAccount sa: storage account
        :param object entity: Storage File or Blob object
        :param bool is_file: is a file object
        :param str container: container
        :param str dir: Azure File directory structure
        :rtype: StorageEntity
        :return: Azure storage entity object
        """
        vio = blobxfer.models.metadata.vectored_io_from_metadata(
            entity.metadata)
        if not isinstance(vio, blobxfer.models.metadata.VectoredStripe):
            ase = self._convert_to_storage_entity_with_encryption_metadata(
                options, store_raw_metadata, sa, entity, None, is_file,
                container, dir, file_snapshot)
            yield ase
            return
        # if this slice is not the first, ignore. the reason for this is
        # 1. Ensures direct get on a slice does nothing unless the
        # zero-th blob is retrieved/accessed (eliminates partial data
        # download), which will reconstruct all of the stripes via next
        # pointers
        # 2. Data is not retrieved multiple times for the same slice without
        # having to maintain a fetched map
        if vio.slice_id != 0:
            yield None
            return
        # yield this entity
        ase = self._convert_to_storage_entity_with_encryption_metadata(
            options, store_raw_metadata, sa, entity, vio, is_file, container,
            dir, file_snapshot)
        yield ase
        # iterate all slices
        while vio.next is not None:
            # follow next pointer
            sa = creds.get_storage_account(vio.next.storage_account_name)
            if is_file:
                entity = blobxfer.operations.azure.file.get_file_properties(
                    sa.file_client, vio.next.container, vio.next.name,
                    snapshot=file_snapshot)
                _, dir = blobxfer.util.explode_azure_path(vio.next.name)
            else:
                entity = blobxfer.operations.azure.blob.get_blob_properties(
                    sa.block_blob_client, vio.next.container, vio.next.name,
                    ase.mode)
            vio = blobxfer.models.metadata.vectored_io_from_metadata(
                entity.metadata)
            # yield next
            ase = self._convert_to_storage_entity_with_encryption_metadata(
                options, store_raw_metadata, sa, entity, vio, is_file,
                container, dir, file_snapshot)
            yield ase

    def _populate_from_list_files(self, creds, options, dry_run):
        # type: (SourcePath, StorageCredentials, Any, bool) -> StorageEntity
        """Internal generator for Azure remote files
        :param SourcePath self: this
        :param StorageCredentials creds: storage creds
        :param object options: download or synccopy options
        :param bool dry_run: dry run
        :rtype: StorageEntity
        :return: Azure storage entity object
        """
        store_raw_metadata = isinstance(
            options, blobxfer.models.options.SyncCopy)
        for _path in self._paths:
            rpath = str(_path)
            sa = creds.get_storage_account(self.lookup_storage_account(rpath))
            # ensure at least read permissions
            if not sa.can_read_object:
                raise RuntimeError(
                    'unable to populate sources for remote path {} as '
                    'credential for storage account {} does not permit read '
                    'access'.format(rpath, sa.name))
            cont, dir = blobxfer.util.explode_azure_path(rpath)
            snapshot = None
            if dir is not None:
                # parse out snapshot if part of url
                a, b, snapshot = \
                    blobxfer.operations.azure.file.parse_file_path(dir)
                # check if "dir" is a file or a vdir
                chk = blobxfer.operations.azure.file.check_if_single_file(
                    sa.file_client, cont, dir)
                if chk[0]:
                    dir = a
                else:
                    if blobxfer.util.is_not_empty(a):
                        dir = '/'.join((a, b))
                    else:
                        dir = b
            if snapshot is None:
                _, cont, snapshot = \
                    blobxfer.operations.azure.file.parse_file_path(cont)
            if sa.can_list_container_objects:
                for file in blobxfer.operations.azure.file.list_files(
                        sa.file_client, cont, dir, options.recursive,
                        snapshot=snapshot):
                    if not self._inclusion_check(file.name):
                        if dry_run:
                            logger.info(
                                '[DRY RUN] skipping due to filters: '
                                '{}/{}'.format(cont, file.name))
                        continue
                    for ase in self._handle_vectored_io_stripe(
                            creds, options, store_raw_metadata, sa, file, True,
                            cont, dir=None, file_snapshot=snapshot):
                        if ase is None:
                            continue
                        yield ase
            else:
                file = blobxfer.operations.azure.file.get_file_properties(
                    sa.file_client, cont, dir, snapshot=snapshot)
                if file is None:
                    logger.error(
                        'file {} not found in storage account {}'.format(
                            rpath, sa.name))
                    return
                if not self._inclusion_check(file.name):
                    if dry_run:
                        logger.info(
                            '[DRY RUN] skipping due to filters: {}/{}'.format(
                                cont, file.name))
                    return
                for ase in self._handle_vectored_io_stripe(
                        creds, options, store_raw_metadata, sa, file, True,
                        cont, dir=None, file_snapshot=snapshot):
                    if ase is None:
                        continue
                    yield ase

    def _populate_from_arbitrary_url(self, remote_path):
        # type: (SourcePath, str) -> StorageEntity
        """Internal generator for Azure remote blobs
        :param SourcePath self: this
        :param str remote_path: remote path
        :rtype: StorageEntity
        :return: Azure storage entity object
        """
        # HEAD remote path to retrieve length
        response = requests.head(remote_path)
        ase = blobxfer.models.azure.StorageEntity(container=None)
        ase.populate_from_arbitrary_url(
            remote_path, int(response.headers['Content-Length']))
        return ase

    def _populate_from_list_blobs(self, creds, options, dry_run):
        # type: (SourcePath, StorageCredentials, Any, bool) -> StorageEntity
        """Internal generator for Azure remote blobs
        :param SourcePath self: this
        :param StorageCredentials creds: storage creds
        :param object options: download or synccopy options
        :param bool dry_run: dry run
        :rtype: StorageEntity
        :return: Azure storage entity object
        """
        is_synccopy = isinstance(options, blobxfer.models.options.SyncCopy)
        for _path in self._paths:
            rpath = str(_path)
            if (is_synccopy and
                    (rpath.lower().startswith('http://') or
                     rpath.lower().startswith('https://'))):
                ase = self._populate_from_arbitrary_url(rpath)
                yield ase
                continue
            sa = creds.get_storage_account(self.lookup_storage_account(rpath))
            # ensure at least read permissions
            if not sa.can_read_object:
                raise RuntimeError(
                    'unable to populate sources for remote path {} as '
                    'credential for storage account {} does not permit read '
                    'access'.format(rpath, sa.name))
            cont, dir = blobxfer.util.explode_azure_path(rpath)
            if sa.can_list_container_objects:
                for blob in blobxfer.operations.azure.blob.list_blobs(
                        sa.block_blob_client, cont, dir, options.mode,
                        options.recursive):
                    # check for virtual directory placeholder
                    if not is_synccopy:
                        try:
                            if (blob.metadata[
                                    _METADATA_VIRTUAL_DIRECTORY] == 'true'):
                                continue
                        except KeyError:
                            pass
                    if not self._inclusion_check(blob.name):
                        if dry_run:
                            logger.info(
                                '[DRY RUN] skipping due to filters: '
                                '{}/{}'.format(cont, blob.name))
                        continue
                    for ase in self._handle_vectored_io_stripe(
                            creds, options, is_synccopy, sa, blob,
                            False, cont):
                        if ase is None:
                            continue
                        yield ase
            else:
                blob = blobxfer.operations.azure.blob.get_blob_properties(
                    sa.block_blob_client, cont, dir, options.mode)
                if blob is None:
                    logger.error(
                        'blob {} not found in storage account {}'.format(
                            rpath, sa.name))
                    return
                if not self._inclusion_check(blob.name):
                    if dry_run:
                        logger.info(
                            '[DRY RUN] skipping due to filters: {}/{}'.format(
                                cont, blob.name))
                    return
                for ase in self._handle_vectored_io_stripe(
                        creds, options, is_synccopy, sa, blob, False, cont):
                    if ase is None:
                        continue
                    yield ase


class DestinationPath(blobxfer.models._BaseSourcePaths):
    """Azure Destination Path"""
    def __init__(self):
        # type: (SourcePath) -> None
        """Ctor for SourcePath
        :param SourcePath self: this
        """
        super().__init__()
        self._path_map = {}

    def add_path_with_storage_account(self, remote_path, storage_account):
        # type: (SourcePath, str, str) -> None
        """Add a path with an associated storage account
        :param SourcePath self: this
        :param str remote_path: remote path
        :param str storage_account: storage account to associate with path
        """
        if len(self._path_map) >= 1:
            raise RuntimeError(
                'cannot add multiple remote paths to SourcePath objects')
        rpath = blobxfer.util.normalize_azure_path(remote_path)
        self.add_path(rpath)
        self._path_map[rpath] = storage_account

    def lookup_storage_account(self, remote_path):
        # type: (SourcePath, str) -> str
        """Lookup the storage account associated with the remote path
        :param SourcePath self: this
        :param str remote_path: remote path
        :rtype: str
        :return: storage account associated with path
        """
        return self._path_map[blobxfer.util.normalize_azure_path(remote_path)]



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import collections
import logging
import math
import os
try:
    import pathlib2 as pathlib
except ImportError:  # noqa
    import pathlib
import tempfile
import threading
import time
# non-stdlib imports
# local imports
import blobxfer.models.azure
import blobxfer.models.crypto
import blobxfer.models.options
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)
# global defines
_AUTO_SELECT_CHUNKSIZE_BYTES = 8388608
# named tuples
Offsets = collections.namedtuple(
    'Offsets', [
        'chunk_num',
        'fd_start',
        'num_bytes',
        'range_end',
        'range_start',
        'unpad',
    ]
)
UncheckedChunk = collections.namedtuple(
    'UncheckedChunk', [
        'data_len',
        'fd_start',
        'file_path',
        'temp',
    ]
)
LocalPathView = collections.namedtuple(
    'LocalPathView', [
        'fd_end',
        'fd_start',
    ]
)


class LocalDestinationPath(object):
    """Local Destination Path"""
    def __init__(self, path=None):
        # type: (LocalDestinationPath, str) -> None
        """Ctor for LocalDestinationPath
        :param LocalDestinationPath self: this
        :param str path: path
        """
        self._is_dir = None
        if path is not None:
            self.path = path

    @property
    def path(self):
        # type: (LocalDestinationPath) -> pathlib.Path
        """Path property
        :param LocalDestinationPath self: this
        :rtype: pathlib.Path
        :return: local destination path
        """
        return self._path

    @path.setter
    def path(self, value):
        # type: (LocalDestinationPath, str) -> None
        """Path property setter
        :param LocalDestinationPath self: this
        :param str value: value to set path to
        """
        self._path = pathlib.Path(value)

    @property
    def is_dir(self):
        # type: (LocalDestinationPath) -> bool
        """is_dir property
        :param LocalDestinationPath self: this
        :rtype: bool
        :return: if local destination path is a directory
        """
        return self._is_dir

    @is_dir.setter
    def is_dir(self, value):
        # type: (LocalDestinationPath, bool) -> None
        """is_dir property setter
        :param LocalDestinationPath self: this
        :param bool value: value to set is_dir to
        """
        self._is_dir = value

    def ensure_path_exists(self):
        # type: (LocalDestinationPath) -> None
        """Ensure path exists
        :param LocalDestinationPath self: this
        """
        if self._is_dir is None:
            raise RuntimeError('is_dir not set')
        if self._is_dir:
            self._path.mkdir(mode=0o750, parents=True, exist_ok=True)
        else:
            if self._path.exists() and self._path.is_dir():
                raise RuntimeError(
                    ('destination path {} already exists and is a '
                     'directory').format(self._path))
            else:
                # ensure parent path exists and is created
                self._path.parent.mkdir(
                    mode=0o750, parents=True, exist_ok=True)


class Specification(object):
    """Download Specification"""
    def __init__(
            self, download_options, skip_on_options, local_destination_path):
        # type: (Specification, blobxfer.models.options.Download,
        #        blobxfer.models.options.SkipOn, LocalDestinationPath) -> None
        """Ctor for Specification
        :param DownloadSpecification self: this
        :param blobxfer.models.options.Download download_options:
            download options
        :param blobxfer.models.options.SkipOn skip_on_options: skip on options
        :param LocalDestinationPath local_destination_path: local dest path
        """
        self.options = download_options
        self.skip_on = skip_on_options
        self.destination = local_destination_path
        self.sources = []
        # validate compatible options
        if not self.options.check_file_md5 and self.skip_on.md5_match:
            raise ValueError(
                'cannot specify skip on MD5 match without file MD5 enabled')
        if (self.options.restore_file_properties.attributes and
                not blobxfer.util.on_windows() and os.getuid() != 0):
            logger.warning('cannot set file uid/gid without root privileges')
        if self.options.chunk_size_bytes < 0:
            raise ValueError('chunk size cannot be negative')
        if self.options.max_single_object_concurrency < 1:
            raise ValueError(
                'must specify a positive value for max single object '
                'concurrency')

    def add_azure_source_path(self, source):
        # type: (Specification, blobxfer.operations.azure.SourcePath) -> None
        """Add an Azure Source Path
        :param DownloadSpecification self: this
        :param blobxfer.operations.Azure.SourcePath source:
            Azure source path to add
        """
        self.sources.append(source)


class Descriptor(object):
    """Download Descriptor"""

    _AES_BLOCKSIZE = blobxfer.models.crypto.AES256_BLOCKSIZE_BYTES

    def __init__(self, lpath, ase, options, general_options, resume_mgr):
        # type: (Descriptor, pathlib.Path,
        #        blobxfer.models.azure.StorageEntity,
        #        blobxfer.models.options.Download,
        #        blobxfer.models.options.General,
        #        blobxfer.operations.resume.DownloadResumeManager) -> None
        """Ctor for Descriptor
        :param Descriptor self: this
        :param pathlib.Path lpath: local path
        :param blobxfer.models.azure.StorageEntity ase: Azure Storage Entity
        :param blobxfer.models.options.Download options: download options
        :param blobxfer.models.options.General general_options: general options
        :param blobxfer.operations.resume.DownloadResumeManager resume_mgr:
            download resume manager
        """
        self._verbose = general_options.verbose
        self._offset = 0
        self._chunk_num = 0
        self._next_integrity_chunk = 0
        self._unchecked_chunks = {}
        self._allocated = False
        self._finalized = False
        self._meta_lock = threading.Lock()
        self._hasher_lock = threading.Lock()
        self._resume_mgr = resume_mgr
        self._restore_file_properties = options.restore_file_properties
        self._ase = ase
        # set path
        self.final_path = lpath
        self.view = None
        # auto-select chunk size
        if options.chunk_size_bytes == 0:
            chunk_size_bytes = _AUTO_SELECT_CHUNKSIZE_BYTES
        else:
            chunk_size_bytes = options.chunk_size_bytes
        self._chunk_size = min((chunk_size_bytes, self._ase.size))
        # calculate the total number of ops required for transfer
        self._total_chunks = self._compute_total_chunks(self._chunk_size)
        self._outstanding_ops = self._total_chunks
        # initialize integrity checkers
        self.hmac = None
        self.md5 = None
        self._integrity_failed = False
        self._initialize_integrity_checkers(options)

    @property
    def entity(self):
        # type: (Descriptor) -> blobxfer.models.azure.StorageEntity
        """Get linked blobxfer.models.azure.StorageEntity
        :param Descriptor self: this
        :rtype: blobxfer.models.azure.StorageEntity
        :return: blobxfer.models.azure.StorageEntity
        """
        return self._ase

    @property
    def must_compute_md5(self):
        # type: (Descriptor) -> bool
        """Check if MD5 must be computed
        :param Descriptor self: this
        :rtype: bool
        :return: if MD5 must be computed
        """
        return self.md5 is not None

    @property
    def all_operations_completed(self):
        # type: (Descriptor) -> bool
        """All operations are completed
        :param Descriptor self: this
        :rtype: bool
        :return: if all operations completed
        """
        with self._meta_lock:
            return (self._outstanding_ops == 0 and
                    len(self._unchecked_chunks) == 0)

    @property
    def is_resumable(self):
        # type: (Descriptor) -> bool
        """Download is resume capable
        :param Descriptor self: this
        :rtype: bool
        :return: if resumable
        """
        return self._resume_mgr is not None and self.hmac is None

    def _compute_total_chunks(self, chunk_size):
        # type: (Descriptor, int) -> int
        """Compute total number of chunks for entity
        :param Descriptor self: this
        :param int chunk_size: chunk size
        :rtype: int
        :return: num chunks
        """
        try:
            return int(math.ceil(self._ase.size / chunk_size))
        except ZeroDivisionError:
            return 0

    def _initialize_integrity_checkers(self, options):
        # type: (Descriptor, blobxfer.models.options.Download) -> None
        """Initialize file integrity checkers
        :param Descriptor self: this
        :param blobxfer.models.options.Download options: download options
        """
        if self._ase.is_encrypted:
            # ensure symmetric key exists
            if blobxfer.util.is_none_or_empty(
                    self._ase.encryption_metadata.symmetric_key):
                raise RuntimeError(
                    'symmetric key is invalid: provide RSA private key '
                    'or metadata corrupt')
            self.hmac = self._ase.encryption_metadata.initialize_hmac()
        if (self.hmac is None and options.check_file_md5 and
                blobxfer.util.is_not_empty(self._ase.md5)):
            self.md5 = blobxfer.util.new_md5_hasher()

    @staticmethod
    def compute_allocated_size(size, is_encrypted):
        # type: (int, bool) -> int
        """Compute allocated size on disk
        :param int size: size (content length)
        :param bool is_ecrypted: if entity is encrypted
        :rtype: int
        :return: required size on disk
        """
        # compute size
        if size > 0:
            if is_encrypted:
                # cipher_len_without_iv = (clear_len / aes_bs + 1) * aes_bs
                allocatesize = (
                    size //
                    blobxfer.models.download.Descriptor._AES_BLOCKSIZE - 1
                ) * blobxfer.models.download.Descriptor._AES_BLOCKSIZE
                if allocatesize < 0:
                    raise RuntimeError('allocatesize is negative')
            else:
                allocatesize = size
        else:
            allocatesize = 0
        return allocatesize

    @staticmethod
    def generate_view(ase):
        # type: (blobxfer.models.azure.StorageEntity) ->
        #       Tuple[LocalPathView, int]
        """Generate local path view and total size required
        :param blobxfer.models.azure.StorageEntity ase: Storage Entity
        :rtype: tuple
        :return: (local path view, allocation size)
        """
        slicesize = blobxfer.models.download.Descriptor.compute_allocated_size(
            ase.size, ase.is_encrypted)
        if ase.vectored_io is None:
            view = LocalPathView(
                fd_start=0,
                fd_end=slicesize,
            )
            total_size = slicesize
        else:
            view = LocalPathView(
                fd_start=ase.vectored_io.offset_start,
                fd_end=ase.vectored_io.offset_start + slicesize,
            )
            total_size = ase.vectored_io.total_size
        return view, total_size

    @staticmethod
    def convert_vectored_io_slice_to_final_path_name(local_path, ase):
        # type: (pathlib.Path,
        #        blobxfer.models.azure.StorageEntity) -> pathlib.Path
        """Convert vectored io slice to final path name
        :param pathlib.Path local_path: local path
        :param blobxfer.models.azure.StorageEntity ase: Storage Entity
        :rtype: pathlib.Path
        :return: converted final path
        """
        name = blobxfer.models.metadata.\
            remove_vectored_io_slice_suffix_from_name(
                local_path.name, ase.vectored_io.slice_id)
        _tmp = list(local_path.parts[:-1])
        _tmp.append(name)
        return pathlib.Path(*_tmp)

    def _set_final_path_view(self):
        # type: (Descriptor) -> int
        """Set final path view and return required space on disk
        :param Descriptor self: this
        :rtype: int
        :return: required size on disk
        """
        # set final path if vectored io stripe
        if self._ase.vectored_io is not None:
            self.final_path = blobxfer.models.download.Descriptor.\
                convert_vectored_io_slice_to_final_path_name(
                    self.final_path, self._ase)
        # generate view
        view, total_size = blobxfer.models.download.Descriptor.generate_view(
            self._ase)
        self.view = view
        return total_size

    def _allocate_disk_space(self):
        # type: (Descriptor) -> None
        """Perform file allocation (possibly sparse)
        :param Descriptor self: this
        """
        with self._meta_lock:
            if self._allocated or self._offset != 0:
                return
            # set local path view
            allocatesize = self._set_final_path_view()
            # check if path already exists and is of sufficient size
            if (not self.final_path.exists() or
                    self.final_path.stat().st_size != allocatesize):
                # create parent path
                self.final_path.parent.mkdir(
                    mode=0o750, parents=True, exist_ok=True)
                # allocate file
                with self.final_path.open('wb') as fd:
                    if allocatesize > 0:
                        try:
                            os.posix_fallocate(fd.fileno(), 0, allocatesize)
                        except AttributeError:
                            fd.seek(allocatesize - 1)
                            fd.write(b'\0')
            self._allocated = True

    def _resume(self):
        # type: (Descriptor) -> int
        """Resume a download, if possible
        :param Descriptor self: this
        :rtype: int or None
        :return: verified download offset
        """
        if self._resume_mgr is None or self._offset > 0 or self._finalized:
            return None
        # check if path exists in resume db
        rr = self._resume_mgr.get_record(self._ase)
        if rr is None:
            logger.debug('no resume record for {}'.format(self.final_path))
            return None
        # ensure lengths are the same
        if rr.length != self._ase.size:
            logger.warning('resume length mismatch {} -> {}'.format(
                rr.length, self._ase.size))
            return None
        # calculate current chunk and offset
        if rr.next_integrity_chunk == 0:
            logger.debug('nothing to resume for {}'.format(self.final_path))
            return None
        curr_chunk = rr.next_integrity_chunk
        # set offsets if completed and the final path exists
        if rr.completed and self.final_path.exists():
            with self._meta_lock:
                logger.debug('{} download already completed'.format(
                    self.final_path))
                self._offset = self._ase.size
                self._chunk_num = curr_chunk
                self._chunk_size = rr.chunk_size
                self._total_chunks = self._compute_total_chunks(rr.chunk_size)
                self._next_integrity_chunk = rr.next_integrity_chunk
                self._outstanding_ops = 0
                self._finalized = True
            return self._ase.size
        # encrypted files are not resumable due to hmac requirement
        if self._ase.is_encrypted:
            logger.debug('cannot resume encrypted entity {}'.format(
                self._ase.path))
            return None
        self._allocate_disk_space()
        # check if final path exists
        if not self.final_path.exists():  # noqa
            logger.warning('download path {} does not exist'.format(
                self.final_path))
            return None
        if self.hmac is not None:
            raise RuntimeError(
                'unexpected hmac object for entity {}'.format(self._ase.path))
        # re-hash from 0 to offset if needed
        _fd_offset = 0
        _end_offset = min((curr_chunk * rr.chunk_size, rr.length))
        if self.md5 is not None and curr_chunk > 0:
            _blocksize = blobxfer.util.MEGABYTE << 2
            logger.debug(
                'integrity checking existing file {} offset {} -> {}'.format(
                    self.final_path,
                    self.view.fd_start,
                    self.view.fd_start + _end_offset)
            )
            with self._hasher_lock:
                with self.final_path.open('rb') as filedesc:
                    filedesc.seek(self.view.fd_start, 0)
                    while _fd_offset < _end_offset:
                        if (_fd_offset + _blocksize) > _end_offset:
                            _blocksize = _end_offset - _fd_offset
                        _buf = filedesc.read(_blocksize)
                        self.md5.update(_buf)
                        _fd_offset += _blocksize
            del _blocksize
            # compare hashes
            hexdigest = self.md5.hexdigest()
            if rr.md5hexdigest != hexdigest:
                logger.warning(
                    'MD5 mismatch resume={} computed={} for {}'.format(
                        rr.md5hexdigest, hexdigest, self.final_path))
                # reset hasher
                self.md5 = blobxfer.util.new_md5_hasher()
                return None
        # set values from resume
        with self._meta_lock:
            self._offset = _end_offset
            self._chunk_num = curr_chunk
            self._chunk_size = rr.chunk_size
            self._total_chunks = self._compute_total_chunks(rr.chunk_size)
            self._next_integrity_chunk = rr.next_integrity_chunk
            self._outstanding_ops = (
                self._total_chunks - self._next_integrity_chunk
            )
            logger.debug(
                ('resuming file {} from byte={} chunk={} chunk_size={} '
                 'total_chunks={} next_integrity_chunk={} '
                 'outstanding_ops={}').format(
                     self.final_path, self._offset, self._chunk_num,
                     self._chunk_size, self._total_chunks,
                     self._next_integrity_chunk, self._outstanding_ops))
        return _end_offset

    def cleanup_all_temporary_files(self):
        # type: (Descriptor) -> None
        """Cleanup all temporary files in case of an exception or interrupt.
        This function is not thread-safe.
        :param Descriptor self: this
        """
        # delete local file
        try:
            self.final_path.unlink()
        except OSError:
            pass
        # iterate unchecked chunks and delete
        for key in self._unchecked_chunks:
            ucc = self._unchecked_chunks[key]['ucc']
            if ucc.temp:
                try:
                    ucc.file_path.unlink()
                except OSError:
                    pass

    def next_offsets(self):
        # type: (Descriptor) -> Offsets
        """Retrieve the next offsets
        :param Descriptor self: this
        :rtype: Offsets
        :return: download offsets
        """
        resume_bytes = self._resume()
        if resume_bytes is None and not self._allocated:
            self._allocate_disk_space()
        with self._meta_lock:
            if self._offset >= self._ase.size:
                return None, resume_bytes
            if self._offset + self._chunk_size > self._ase.size:
                chunk = self._ase.size - self._offset
            else:
                chunk = self._chunk_size
            # on download, num_bytes must be offset by -1 as the x-ms-range
            # header expects it that way. x -> y bytes means first bits of the
            # (x+1)th byte to the last bits of the (y+1)th byte. for example,
            # 0 -> 511 means byte 1 to byte 512
            num_bytes = chunk - 1
            chunk_num = self._chunk_num
            fd_start = self._offset
            range_start = self._offset
            if self._ase.is_encrypted:
                # ensure start is AES block size aligned
                range_start = range_start - \
                    (range_start % self._AES_BLOCKSIZE) - \
                    self._AES_BLOCKSIZE
                if range_start <= 0:
                    range_start = 0
            range_end = self._offset + num_bytes
            self._offset += chunk
            self._chunk_num += 1
            if self._ase.is_encrypted and self._offset >= self._ase.size:
                unpad = True
            else:
                unpad = False
            return Offsets(
                chunk_num=chunk_num,
                fd_start=fd_start,
                num_bytes=chunk,
                range_start=range_start,
                range_end=range_end,
                unpad=unpad,
            ), resume_bytes

    def hmac_iv(self, iv):
        # type: (Descriptor, bytes) -> None
        """Send IV through hasher
        :param Descriptor self: this
        :param bytes iv: iv
        """
        with self._hasher_lock:
            self.hmac.update(iv)

    def write_unchecked_data(self, offsets, data):
        # type: (Descriptor, Offsets, bytes) -> None
        """Write unchecked data to disk
        :param Descriptor self: this
        :param Offsets offsets: download offsets
        :param bytes data: data
        """
        self.write_data(offsets, data)
        unchecked = UncheckedChunk(
            data_len=len(data),
            fd_start=self.view.fd_start + offsets.fd_start,
            file_path=self.final_path,
            temp=False,
        )
        with self._meta_lock:
            self._unchecked_chunks[offsets.chunk_num] = {
                'ucc': unchecked,
                'decrypted': True,
            }

    def write_unchecked_hmac_data(self, offsets, data):
        # type: (Descriptor, Offsets, bytes) -> None
        """Write unchecked encrypted data to disk
        :param Descriptor self: this
        :param Offsets offsets: download offsets
        :param bytes data: hmac/encrypted data
        """
        fname = None
        with tempfile.NamedTemporaryFile(mode='wb', delete=False) as fd:
            fname = fd.name
            fd.write(data)
        unchecked = UncheckedChunk(
            data_len=len(data),
            fd_start=0,
            file_path=pathlib.Path(fname),
            temp=True,
        )
        with self._meta_lock:
            self._unchecked_chunks[offsets.chunk_num] = {
                'ucc': unchecked,
                'decrypted': False,
            }
        return str(unchecked.file_path)

    def mark_unchecked_chunk_decrypted(self, chunk_num):
        # type: (Descriptor, int) -> None
        """Mark an unchecked chunk as decrypted
        :param Descriptor self: this
        :param int chunk_num: unchecked chunk number
        """
        with self._meta_lock:
            self._unchecked_chunks[chunk_num]['decrypted'] = True

    def perform_chunked_integrity_check(self):
        # type: (Descriptor) -> None
        """Hash data against stored hasher safely
        :param Descriptor self: this
        """
        hasher = self.hmac or self.md5
        # iterate from next chunk to be checked
        while True:
            ucc = None
            with self._meta_lock:
                chunk_num = self._next_integrity_chunk
                # check if the next chunk is ready
                if (chunk_num in self._unchecked_chunks and
                        self._unchecked_chunks[chunk_num]['decrypted']):
                    ucc = self._unchecked_chunks.pop(chunk_num)['ucc']
                else:
                    break
            # hash data and set next integrity chunk
            md5hexdigest = None
            if hasher is not None:
                with ucc.file_path.open('rb') as fd:
                    if not ucc.temp:
                        fd.seek(ucc.fd_start, 0)
                    chunk = fd.read(ucc.data_len)
                if ucc.temp:
                    ucc.file_path.unlink()
                with self._hasher_lock:
                    hasher.update(chunk)
                    if hasher == self.md5:
                        md5hexdigest = hasher.hexdigest()
            with self._meta_lock:
                # update integrity counter and resume db
                self._next_integrity_chunk += 1
                if self.is_resumable:
                    self._resume_mgr.add_or_update_record(
                        self.final_path, self._ase, self._chunk_size,
                        self._next_integrity_chunk, False, md5hexdigest,
                    )
                # decrement outstanding op counter
                self._outstanding_ops -= 1

    def _update_resume_for_completed(self):
        # type: (Descriptor) -> None
        """Update resume for completion
        :param Descriptor self: this
        """
        if not self.is_resumable:
            return
        with self._meta_lock:
            self._resume_mgr.add_or_update_record(
                self.final_path, self._ase, self._chunk_size,
                self._next_integrity_chunk, True, None,
            )

    def write_data(self, offsets, data):
        # type: (Descriptor, Offsets, bytes) -> None
        """Write data to disk
        :param Descriptor self: this
        :param Offsets offsets: download offsets
        :param bytes data: data
        """
        if len(data) > 0:
            # offset from internal view
            pos = self.view.fd_start + offsets.fd_start
            with self.final_path.open('r+b') as fd:
                fd.seek(pos, 0)
                fd.write(data)

    def finalize_integrity(self):
        # type: (Descriptor) -> None
        """Finalize integrity check for download
        :param Descriptor self: this
        """
        with self._meta_lock:
            if self._finalized:
                return
        # check final file integrity
        check = False
        msg = None
        if self.hmac is not None:
            mac = self._ase.encryption_metadata.encryption_authentication.\
                message_authentication_code
            digest = blobxfer.util.base64_encode_as_string(self.hmac.digest())
            if digest == mac:
                check = True
            msg = '{}: {}, {} {} <L..R> {}'.format(
                self._ase.encryption_metadata.encryption_authentication.
                algorithm,
                'OK' if check else 'MISMATCH',
                self._ase.path,
                digest,
                mac,
            )
        elif self.md5 is not None:
            digest = blobxfer.util.base64_encode_as_string(self.md5.digest())
            if digest == self._ase.md5:
                check = True
            msg = 'MD5: {}, {} {} <L..R> {}'.format(
                'OK' if check else 'MISMATCH',
                self._ase.path,
                digest,
                self._ase.md5,
            )
        else:
            check = True
            msg = 'MD5: SKIPPED, {} None <L..R> {}'.format(
                self._ase.path,
                self._ase.md5
            )
        # cleanup if download failed
        if not check:
            self._integrity_failed = True
            logger.error(msg)
        else:
            logger.info(msg)

    def _restore_file_attributes(self):
        # type: (Descriptor) -> None
        """Restore file attributes for file
        :param Descriptor self: this
        """
        if (not self._restore_file_properties.attributes or
                self._ase.file_attributes is None):
            return
        # set file uid/gid and mode
        if blobxfer.util.on_windows():  # noqa
            # TODO not implemented yet
            pass
        else:
            self.final_path.chmod(int(self._ase.file_attributes.mode, 8))
            if os.getuid() == 0:  # noqa
                os.chown(
                    str(self.final_path),
                    self._ase.file_attributes.uid,
                    self._ase.file_attributes.gid
                )

    def _restore_file_lmt(self):
        # type: (Descriptor) -> None
        """Restore file lmt for file
        :param Descriptor self: this
        """
        if not self._restore_file_properties.lmt or self._ase.lmt is None:
            return
        # timestamp() func is not available in py27
        ts = time.mktime(self._ase.lmt.timetuple())
        os.utime(str(self.final_path), (ts, ts))

    def finalize_file(self):
        # type: (Descriptor) -> None
        """Finalize file for download
        :param Descriptor self: this
        """
        # delete bad file if integrity failed
        if self._integrity_failed:
            self.final_path.unlink()
        else:
            self._restore_file_attributes()
            self._restore_file_lmt()
        # update resume file
        self._update_resume_for_completed()
        with self._meta_lock:
            self._finalized = True



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import errno
import ssl
# non-stdlib imports
import azure.storage.common.models
import azure.storage.common.retry
import requests
import urllib3
# local imports


# global defines
_RETRYABLE_ERRNO_MAXRETRY = frozenset((
    '[Errno {}]'.format(errno.ECONNABORTED),
    '[Errno {}]'.format(errno.ECONNRESET),
    '[Errno {}]'.format(errno.ECONNREFUSED),
    '[Errno {}]'.format(errno.ENETRESET),
    '[Errno {}]'.format(errno.ETIMEDOUT),
    # Windows errors
    '[WinError {}]'.format(errno.ECONNABORTED),
    '[WinError {}]'.format(errno.ECONNRESET),
    '[WinError {}]'.format(errno.ECONNREFUSED),
    '[WinError {}]'.format(errno.ENETRESET),
    '[WinError {}]'.format(errno.ETIMEDOUT),
))
_RETRYABLE_ERRNO_PROTOCOL = frozenset((
    '({},'.format(errno.ECONNABORTED),
    '({},'.format(errno.ECONNRESET),
    '({},'.format(errno.ECONNREFUSED),
    '({},'.format(errno.ENETRESET),
    '({},'.format(errno.ETIMEDOUT),
))
_RETRYABLE_STRING_FALLBACK = frozenset((
    'connection aborted',
    'connection reset',
    'connection refused',
    'network dropped',
    'timed out',
))


class ExponentialRetryWithMaxWait(azure.storage.common.retry._Retry):
    """Exponential Retry with Max Wait Reset"""
    def __init__(
            self, initial_backoff=0.1, max_backoff=1, max_retries=None,
            reset_at_max=True):
        # type: (ExponentialRetryWithMaxWait, int, int, int, bool) -> None
        """Ctor for ExponentialRetryWithMaxWait
        :param ExponentialRetryWithMaxWait self: this
        :param int initial_backoff: initial backoff
        :param int max_backoff: max backoff
        :param int max_retries: max retries
        :param bool reset_at_max: reset after reaching max wait
        """
        if max_backoff <= 0:
            raise ValueError(
                'max backoff is non-positive: {}'.format(max_backoff))
        if max_retries is not None and max_retries < 0:
            raise ValueError(
                'max retries is invalid: {}'.format(max_retries))
        if max_backoff < initial_backoff:
            raise ValueError(
                'max backoff {} less than initial backoff {}'.format(
                    max_backoff, initial_backoff))
        self._backoff_count = 0
        self._last_backoff = initial_backoff
        self.initial_backoff = initial_backoff
        self.max_backoff = max_backoff
        self.reset_at_max = reset_at_max
        super(ExponentialRetryWithMaxWait, self).__init__(
            max_retries if max_retries is not None else 2147483647, False)

    def _should_retry(self, context):
        # type: (ExponentialRetryWithMaxWait,
        #        azure.storage.common.models.RetryContext) -> bool
        """Determine if retry should happen or not
        :param ExponentialRetryWithMaxWait self: this
        :param azure.storage.common.models.RetryContext context: retry context
        :rtype: bool
        :return: True if retry should happen, False otherwise
        """
        # do not retry if max attempts equal or exceeded
        if context.count >= self.max_attempts:
            return False

        # get response status
        status = None
        if context.response and context.response.status:
            status = context.response.status

        # if there is no response status, then handle the exception
        # appropriately from the lower layer
        if status is None:
            exc = context.exception
            # default to not retry in unknown/unhandled exception case
            ret = False
            # requests timeout, retry
            if isinstance(exc, requests.Timeout):
                ret = True
            elif isinstance(exc, requests.exceptions.ContentDecodingError):
                ret = True
            elif (isinstance(exc, requests.exceptions.ConnectionError) or
                  isinstance(exc, requests.exceptions.ChunkedEncodingError)):
                # newer versions of requests do not expose errno on the
                # args[0] reason object; manually string parse
                if isinstance(exc.args[0], urllib3.exceptions.MaxRetryError):
                    try:
                        msg = exc.args[0].reason.args[0]
                    except (AttributeError, IndexError):
                        # unexpected/malformed exception hierarchy, don't retry
                        pass
                    else:
                        if (isinstance(msg, ssl.SSLWantWriteError) or
                                isinstance(msg, ssl.SSLWantReadError)):
                            ret = True
                        elif any(x in msg for x in _RETRYABLE_ERRNO_MAXRETRY):
                            ret = True
                elif isinstance(exc.args[0], urllib3.exceptions.ProtocolError):
                    try:
                        msg = exc.args[0].args[0]
                    except (AttributeError, IndexError):
                        # unexpected/malformed exception hierarchy, don't retry
                        pass
                    else:
                        if any(x in msg for x in _RETRYABLE_ERRNO_PROTOCOL):
                            ret = True
            # fallback to string search
            if not ret:
                msg = str(exc).lower()
                if any(x in msg for x in _RETRYABLE_STRING_FALLBACK):
                    ret = True
            return ret
        elif 200 <= status < 300:
            # failure during respond body download or parsing, so success
            # codes should be retried
            return True
        elif 300 <= status < 500:
            # response code 404 should be retried if secondary was used
            if (status == 404 and
                    context.location_mode ==
                    azure.storage.common.models.LocationMode.SECONDARY):
                return True
            # response code 408 is a timeout and should be retried
            # response code 429 is too many requests (throttle)
            # TODO use "Retry-After" header for backoff amount
            if status == 408 or status == 429:
                return True
            return False
        elif status >= 500:
            # response codes above 500 should be retried except for
            # 501 (not implemented) and 505 (version not supported)
            if status == 501 or status == 505:
                return False
            return True
        else:  # noqa
            # this should be unreachable, retry anyway
            return True

    def retry(self, context):
        # type: (ExponentialRetryWithMaxWait,
        #        azure.storage.common.models.RetryContext) -> int
        """Retry handler
        :param ExponentialRetryWithMaxWait self: this
        :param azure.storage.common.models.RetryContext context: retry context
        :rtype: int or None
        :return: int
        """
        return self._retry(context, self._backoff)

    def _backoff(self, context):
        # type: (ExponentialRetryWithMaxWait,
        #        azure.storage.common.models.RetryContext) -> int
        """Backoff calculator
        :param ExponentialRetryWithMaxWait self: this
        :param azure.storage.common.models.RetryContext context: retry context
        :rtype: int
        :return: backoff amount
        """
        self._backoff_count += 1
        if self._backoff_count == 1:
            self._last_backoff = self.initial_backoff
        else:
            self._last_backoff *= 2
        if self._last_backoff > self.max_backoff and self.reset_at_max:
            self._backoff_count = 1
            self._last_backoff = self.initial_backoff
        return self._last_backoff
